# Test various formulas for "powersumz"
# using computations in the fractional field associated to the polynomial ring over the variables $x_1,\ldots,x_n$,
# or in the polynomial ring over the variables $z_1,\ldots,z_n$.

def PRing(letter, n):
   names = [letter + str(k) for k in range(n)]
   return PolynomialRing(QQ, names)

def vars(letter, n):
   return PRing(letter, n).gens()

def powersum(x, m):
   n = len(x)
   return sum(x[j] ^ m for j in range(n))

def coef_palpowersum_via_powersum(m, k):
   s = (-1) ^ k
   return (s * m * binomial(m - k, k) / (m - k)) if m > 0 else 2

def powersumz_via_lincomb_powersum(z, m):
   n = len(z)
   F = z[0].parent()
   m1 = floor((m - 1) / 2)
   coefs = [((-1) ** k) * m * binomial(m - k, k) / (m - k) for k in range(m1 + 1)]
   ps = [powersum(z, m - 2 * k) for k in range(m1 + 1)]
   main_sum = sum(coefs[k] * ps[k] for k in range(m1 + 1))
   last_summand = ((-1) ** floor(m / 2)) * 2 * n * F.one() if mod(m, 2) == 0 else F.zero()
   return main_sum + last_summand

def test_powersumz_via_lincomb_powersum_polx(n, m, verbos):
   if verbos >= 2:
      print('test_powersumz_via_lincomb_powersum_polx, n = %d, m = %d' % (n, m))
   x = vars('x', n)
   y = list(x) + [1 / x[j] for j in range(n)]
   z = [x[j] + 1 / x[j] for j in range(n)]
   res0 = powersum(y, m)
   res1 = powersumz_via_lincomb_powersum(z, m)
   if verbos >= 3:
      print('res0 = ' + str(res0))
      print('res1 = ' + str(res1))
   return res0 == res1

def big_test_powersumz_via_lincomb_powersum_polx(nmax, mmax, verbos):
   if verbos >= 1:
      print('big_test_powersumz_via_lincomb_powersum_polx, nmax = %d, mmax = %d' % (nmax, mmax))
   tasks = [(n, m) for n in range(1, nmax + 1) for m in range(mmax + 1)]
   return all(test_powersumz_via_lincomb_powersum_polx(n, m, verbos) for (n, m) in tasks)

def powersumz_via_chebT(z, m):
   return sum(2 * chebyshev_T(m, zelem / 2) for zelem in z)

def test_powersumz_various_formulas_polz(n, m, verbos):
   if verbos >= 2:
      print('test_powersumz_various_formulas_polz, n = %d, m = %d' % (n, m))
   z = vars('z', n)
   res0 = powersumz_via_lincomb_powersum(z, m)
   res1 = powersumz_via_chebT(z, m)
   if verbos >= 3:
      print('res0 = ' + str(res0))
      print('res1 = ' + str(res1))
   return res0 == res1

def big_test_powersumz_various_formulas_polz(nmax, mmax, verbos):
   if verbos >= 1:
      print('big_test_powersumz_various_formulas_polz, nmax = %d, mmax = %d' % (nmax, mmax))
   tasks = [(n, m) for n in range(1, nmax + 1) for m in range(mmax + 1)]
   return all(test_powersumz_various_formulas_polz(n, m, verbos) for (n, m) in tasks)

def coef_powersum_via_palpowersum(m, k):
   return binomial(m, k) if 2 * k < m else binomial(m, k) / 2

def powersum(x, m):
   n = len(x)
   return sum([x[j] ^ m for j in range(n)])

def test_powersum_as_lincomb_powersumz_polx(n, m, verbos):
   if verbos >= 2:
      print('test_powersum_as_lincomb_powersumz_polx, n = %d, m = %d' % (n, m))
   x = vars('x', n)
   y = list(x) + [1 / x[j] for j in range(n)]
   z = [x[j] + 1 / x[j] for j in range(n)]
   F = z[0].parent()
   lhs = powersum(z, m)
   m1 = floor((m - 1) / 2)
   coefs = [binomial(m, k) for k in range(m1 + 1)]
   main_sum = sum(coefs[k] * powersum(y, m - 2 * k) for k in range(m1 + 1))
   last_summand = binomial(m, floor(m / 2)) * n * F.one() if mod(m, 2) == 0 else F.zero()
   rhs = main_sum + last_summand
   if verbos >= 3:
      print('lhs = ' + str(lhs))
      print('rhs = ' + str(rhs))
   return lhs == rhs

def big_test_powersum_as_lincomb_powersumz_polx(nmax, mmax, verbos):
   if verbos >= 1:
      print('big_test_powersum_as_lincomb_powersumz_polx, nmax = %d, mmax = %d' % (nmax, mmax))
   tasks = [(n, m) for n in range(1, nmax + 1) for m in range(mmax + 1)]
   return all(test_powersum_as_lincomb_powersumz_polx(n, m, verbos) for (n, m) in tasks)

import time

def all_tests_powersumz_pol(nmax):
   mmax = 2 * nmax
   verbos = 2
   res = [0] * 3
   t = time.time()
   res[0] = big_test_powersumz_via_lincomb_powersum_polx(nmax, mmax, verbos)
   res[1] = big_test_powersumz_various_formulas_polz(nmax, mmax, verbos)
   res[2] = big_test_powersum_as_lincomb_powersumz_polx(nmax, mmax, verbos)
   total_result = all(res)
   t = time.time() - t
   print('time = %.2f seconds' % t)
   print('total result = ' + str(total_result))

# with nmax = 20, time < 60 sec
all_tests_powersumz_pol(20)

